package rbf;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.text.DecimalFormat;
import figure.Cube3D;
import figure.Point3D;
import figure.Transform;
import figure.Vector3D;
import gui.BasicViewCanvas;

//--------------------------------------------------------------------//
/**
 * \LoX
 */
public class RbfViewCanvas extends BasicViewCanvas{

/** _̕`攼a */
  protected static final int POINT_R=3;

/** _̕`撼a */
  protected static final int POINT_D=2*POINT_R;

/** ̍ŏl */
  public static final double MIN_VALUE=1.0E-10;

/** l */
  protected static final DecimalFormat
    FORM=new DecimalFormat("0.000E0");

/** <i>x</i> iq_ */
  protected int nx;

/** <i>y</i> iq_ */
  protected int ny;

/** <i>z</i> iq_ */
  protected int nz;

/** ˊ֐ԃf[^f */
  protected RbfDataModel model;

/** iq͈ */
  protected Cube3D range;

/** f\̈TCY */
  protected int viewSize;

/** ʒ_ */
  protected Point centerPoint;

/** ϊs */
  protected Transform transform;

/** fsړW */
  protected double transCoef;

/** f]xW */
  protected double rotateCoef;

/** fgEkW */
  protected double zoomCoef;

/** W */
  protected Coordinates coord;

/** G[bZ[W */
  protected String message;

/** tHg */
  protected Font font;

/** wiF */
  protected Color bgColor;

/** OiF */
  protected Color fgColor;

/** _\F */
  protected Color pointColor;

/** iq\F */
  protected Color gridColor;

/**
 * \LoX\z
 * @param model ˊ֐ԃf[^f
 * @param message G[bZ[W
 * @param font tHg
 * @param bgColor wiF
 * @param fgColor OiF
 */
  public RbfViewCanvas(RbfDataModel model,String message,Font font,
      	      	       Color bgColor,Color fgColor){
    this.model=model;
    this.message=message;
    transform=new Transform();
    coord=new Coordinates();
    this.font=font;
    setColors(bgColor,fgColor);
    pointColor=Color.RED;
    gridColor=Color.BLUE;
  }

//--------------------------------------------------------------------//
/**
 * \Fݒ肷
 * @param bgColor wiF
 * @param fgColor OiF
 */
  protected void setColors(Color bgColor,Color fgColor){
    if(bgColor==null){
      this.bgColor=Color.WHITE;
    }
    else{
      this.bgColor=bgColor;
    }
    if(fgColor==null){
      this.fgColor=Color.BLACK;
    }
    else{
      this.fgColor=fgColor;
    }
    setBackground(bgColor);
  }

//--------------------------------------------------------------------//
/**
 * G[bZ[Wݒ肷
 * @param message G[bZ[W
 */
  protected void setMessage(String message){
    this.message=message;
  }

//--------------------------------------------------------------------//
/**
 * ˊ֐ԃf[^fԂ
 * @return ˊ֐ԃf[^f
 */
  public RbfDataModel getModel(){
    return model;
  }

//--------------------------------------------------------------------//
/**
 * f𕽍sړ
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void translate(int dx,int dy){
    transform.translate(transCoef*(double)dx,-transCoef*(double)dy,0.0);
  }

//--------------------------------------------------------------------//
/**
 * f]
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void rotate(int dx,int dy){
    transform.rotate(rotateCoef*(double)dy,rotateCoef*(double)dx,0.0);
  }

//--------------------------------------------------------------------//
/**
 * fgEk
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void zoom(int dx,int dy){
    transform.zoom(Math.pow(2.0,zoomCoef*(double)(dx+dy)));
  }

//--------------------------------------------------------------------//
/**
 * ĕ`悷
 * @param g `悷OtBbNXReLXg
 */
  public void update(Graphics g){
    if(message!=null){
      drawErrorMessage(bg,message);
    }
    else{
      paintModel(bg);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * `悷
 * @param g `悷OtBbNXReLXg
 */
  public void paint(Graphics g){
    setPaint();
    if(message!=null){
      drawErrorMessage(bg,message);
    }
    else{
      paintModel(bg);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * ʐݒ
 */
  protected void setPaint(){
    super.setPaint();
    viewSize=Math.min(panelSize.width,panelSize.height);
    centerPoint=new Point(viewSize/2,viewSize/2);
    transCoef=1.0;
    rotateCoef=Math.PI/(double)viewSize;
    zoomCoef=1.0/(double)viewSize;
  }

//--------------------------------------------------------------------//
/**
 * G[bZ[W\
 * @param g `悷OtBbNXReLXg
 * @param message G[bZ[W
 */
  protected void drawErrorMessage(Graphics2D g,String message){
    g.setColor(fgColor);
    super.drawErrorMessage(g,message);
  }

//--------------------------------------------------------------------//
/**
 * `悷
 * @param g `悷OtBbNXReLXg
 */
  public void paintModel(Graphics2D g){
    g.setColor(bgColor);
    g.fillRect(0,0,panelSize.width,panelSize.height);
    range=model.getRange();

    Point3D center=range.getCenterLocation();
    double boundSize=Math.max(range.width,range.height);
    double vCoef=0.8*(double)viewSize/Math.max(boundSize,MIN_VALUE);
    double csize=0.1*(double)viewSize;
    coord.setVertices(csize,transform);
    Point cc=new Point((int)csize,panelSize.height-(int)csize);

    double cx=centerPoint.getX();
    double cy=centerPoint.getY();
    g.setColor(pointColor);
    int count=model.getPointCount();
    for(int i=0;i<count;i++){
      ViewPoint3D vp=getViewPoint(model.getPoint(i),vCoef,center);
      g.fillOval((int)(cx+vp.x)-POINT_R,(int)(cy-vp.y)-POINT_R,
      	      	 POINT_D,POINT_D);
    }

    g.setColor(gridColor);
    ViewPoint3D vpp[][]=new ViewPoint3D[model.nx+1][model.ny+1];
    for(int i=0;i<=model.nx;i++){
      for(int j=0;j<=model.ny;j++){
      	vpp[i][j]=getViewPoint(model.getGrid(i,j),vCoef,center);
      }
    }
    for(int i=0;i<model.nx;i++){
      for(int j=0;j<=model.ny;j++){
      	g.drawLine((int)(cx+vpp[i][j].x),(int)(cy-vpp[i][j].y),
      	      	   (int)(cx+vpp[i+1][j].x),(int)(cy-vpp[i+1][j].y));
      }
    }
    for(int i=0;i<=model.nx;i++){
      for(int j=0;j<model.ny;j++){
      	g.drawLine((int)(cx+vpp[i][j].x),(int)(cy-vpp[i][j].y),
      	      	   (int)(cx+vpp[i][j+1].x),(int)(cy-vpp[i][j+1].y));
      }
    }
    g.setColor(fgColor);
    coord.paint(g,cc);
  }

//--------------------------------------------------------------------//
/**
 * ʏ̓_Ԃ
 * @param p iq_
 * @param vCoef f̕\{
 * @param center f̒S\ <code>Point3D</code>
 * @return ʏ̓_
 */
  protected ViewPoint3D getViewPoint(Point3D p,double vCoef,
      	      	      	      	     Point3D center){
    ViewPoint3D vp=new ViewPoint3D(vCoef*(p.x-center.x),
      	      	      	      	   vCoef*(p.y-center.y),
      	      	      	      	   vCoef*(p.z-center.z));
    transform.trans(vp);
    return vp;
  }
}
